/**
 *
 * SIROCCO
 * Copyright (C) 2011 France Telecom
 * Contact: sirocco@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  $Id: LibvirtHost.java 394 2011-10-03 13:48:33Z dangtran $
 *
 */

package org.ow2.sirocco.vmm.agent.driver.libvirt;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import javax.management.MBeanNotificationInfo;
import javax.management.Notification;
import javax.management.NotificationEmitter;
import javax.management.ObjectName;

import org.apache.log4j.Logger;
import org.libvirt.Connect;
import org.libvirt.Domain;
import org.libvirt.LibvirtException;
import org.libvirt.NodeInfo;
import org.ow2.sirocco.vmm.agent.domain.AbstractHost;
import org.ow2.sirocco.vmm.agent.driver.util.RemoteExec;
import org.ow2.sirocco.vmm.agent.jmx.MBeanObjectNamer;
import org.ow2.sirocco.vmm.agent.main.AgentCommon;
import org.ow2.sirocco.vmm.api.BadVMPowerStateException;
import org.ow2.sirocco.vmm.api.CustomizationSpec;
import org.ow2.sirocco.vmm.api.DiskOperation;
import org.ow2.sirocco.vmm.api.InsufficientResourcesException;
import org.ow2.sirocco.vmm.api.InvalidVMConfigException;
import org.ow2.sirocco.vmm.api.NotificationTypes;
import org.ow2.sirocco.vmm.api.ResourcePartitionMXBean;
import org.ow2.sirocco.vmm.api.ResourceUsage;
import org.ow2.sirocco.vmm.api.ServerPoolMXBean;
import org.ow2.sirocco.vmm.api.VMConfigSpec;
import org.ow2.sirocco.vmm.api.VMCustomizationSpec;
import org.ow2.sirocco.vmm.api.VMMException;
import org.ow2.sirocco.vmm.api.VNICSpec;
import org.ow2.sirocco.vmm.api.VirtualCdromSpec;
import org.ow2.sirocco.vmm.api.VirtualDiskSpec;
import org.ow2.sirocco.vmm.api.VirtualMachineConfigSpec;
import org.ow2.sirocco.vmm.api.VirtualMachineImageStoreMXBean;
import org.ow2.sirocco.vmm.api.VirtualMachineMXBean;
import org.ow2.sirocco.vmm.api.VirtualMachineMXBean.PowerState;
import org.ow2.sirocco.vmm.api.VnicIPSettings;
import org.ow2.sirocco.vmm.api.VnicIPSettings.IpAssignmentMode;
import org.ow2.sirocco.vmm.api.Volume;

public class LibvirtHost extends AbstractHost implements NotificationEmitter {
    static Logger logger = Logger.getLogger(LibvirtHost.class);

    // the thread pool size determines the maximum number of
    // simultaneous VM creation on the same host
    private static final int MAX_PARALEL_VM_CREATION_PER_HOST = 3;

    private static final int THREADPOOL_SIZE = LibvirtHost.MAX_PARALEL_VM_CREATION_PER_HOST;

    private static ExecutorService executorService = Executors.newFixedThreadPool(LibvirtHost.THREADPOOL_SIZE);

    private String uri;

    private List<LibvirtVirtualMachine> vmList = new CopyOnWriteArrayList<LibvirtVirtualMachine>();

    LibvirtServerPool serverPool;

    private String hostName;

    private HostPowerState hostPowerState = HostPowerState.HALTED;

    private Connect connection;

    private int numCPU = -1;

    private Map<String, String> hostAttributes;

    private RemoteExec.SshAuthInfo authInfo;

    private HostSynchronizer hostSynchronizer;

    String diskImageFormat;

    public static LibvirtHost newHost(final LibvirtServerPool serverPool, final ObjectName objectName, final String hostName,
        final String uri, final RemoteExec.SshAuthInfo authInfo, final Map<String, String> hostAttributes,
        final Boolean isNotStop) {

        try {
            return new LibvirtHost(serverPool, uri, objectName, hostName, authInfo, hostAttributes, isNotStop);
        } catch (LibvirtException ex) {
            LibvirtHost.logger.error("Failed to establish libvirt connection with uri " + uri, ex);
            return null;
        }
    }

    private LibvirtHost(final LibvirtServerPool serverPool, final String uri, final ObjectName objectName,
        final String hostName, final RemoteExec.SshAuthInfo authInfo, final Map<String, String> hostAttributes,
        final Boolean isNotStop) throws LibvirtException {
        super(objectName);
        this.hostPowerState = HostPowerState.HALTED;
        this.uri = uri;
        this.serverPool = serverPool;
        LibvirtHost.logger.debug("Connecting to URI " + this.uri);
        if (isNotStop) {
            try {
                this.connection = this.getLibvirtAPIConnection();
                this.hostPowerState = HostPowerState.RUNNING;
            } catch (LibvirtException ex) {
                LibvirtHost.logger.warn("Host " + hostName + " not responding");
                this.hostPowerState = HostPowerState.NOT_RESPONDING;
            }
        }
        this.hostName = hostName;
        this.authInfo = authInfo;
        this.hostAttributes = new HashMap<String, String>(hostAttributes);
        if (serverPool.getHypervisor().equals("xen")) {
            this.diskImageFormat = LibvirtStoragePool.RAW_FORMAT;
        } else if (serverPool.getHypervisor().equals("kvm")) {
            this.diskImageFormat = LibvirtStoragePool.QCOW2_FORMAT;
        }
        this.hostSynchronizer = new HostSynchronizer(serverPool.getSyncPeriodMillis());
        if (this.hostPowerState == HostPowerState.RUNNING || this.hostPowerState == HostPowerState.NOT_RESPONDING) {
            this.hostSynchronizer.start();
        }
    }

    public boolean isConnectionLost() {
        return this.hostPowerState == HostPowerState.NOT_RESPONDING;
    }

    public RemoteExec.SshAuthInfo getSshAuthInfo() {
        return this.authInfo;
    }

    Connect getLibvirtAPIConnection() throws LibvirtException {
        if (this.connection == null) {
            this.connection = new Connect(this.uri);
        }
        return this.connection;
    }

    List<LibvirtVirtualMachine> getVMs() {
        return this.vmList;
    }

    LibvirtVirtualMachine lookUpVMByName(final String name) {
        for (LibvirtVirtualMachine vm : this.vmList) {
            if (vm.getNameLabel().equals(name)) {
                return vm;
            }
        }
        return null;
    }

    LibvirtVirtualMachine lookUpVMByUUID(final String uuid) {
        for (LibvirtVirtualMachine vm : this.vmList) {
            if (vm.getUuid().equals(uuid)) {
                return vm;
            }
        }
        return null;
    }

    private void addVM(final LibvirtVirtualMachine vm) {
        this.vmList.add(vm);
    }

    private boolean removeVM(final LibvirtVirtualMachine vm) {
        return this.vmList.remove(vm);
    }

    public ServerPoolMXBean getServerPool() {
        return this.serverPool;
    }

    public void postMigrateVM(final LibvirtVirtualMachine vm, final LibvirtHost newHost) {
        this.removeVM(vm);
        newHost.addVM(vm);
    }

    public void removeVMFromInventory(final LibvirtVirtualMachine vm) {
        try {
            AgentCommon.getMBeanServer().unregisterMBean(vm.getObjectName());
        } catch (Exception ex) {
            LibvirtHost.logger.error("Cannot unregister VM MBean", ex);
        }
        vm.release();
        if (this.removeVM(vm)) {
            // this.emitNotification(NotificationTypes.VM_INVENTORY_DEL,
            // "Deleted", vm.getObjectName());
            // LibvirtHost.logger.info("deleted VM " + vm.getObjectName());
        }
    }

    String getIPAddress(String macAddress) throws Exception {
        macAddress = macAddress.toUpperCase();

        LibvirtHost.logger.debug("Determining IP address from MAC address " + macAddress + " ...");

        RemoteExec.Result result = null;
        String command = "getIPfromMAC.sh " + macAddress;

        LibvirtHost.logger.debug("Launching command: " + command);

        try {
            result = RemoteExec.commandAsRoot(this.hostName, this.authInfo, command);
        } catch (RemoteExec.SshException ex) {
            LibvirtHost.logger.debug("SSH failure", ex);
            throw new Exception(ex);
        }

        if (result.exitCode != 0) {
            LibvirtHost.logger.error("Cannot get IP from MAC " + result.output);
            throw new Exception(result.error);
        }

        String ipAddress = result.output;
        if (ipAddress != null && ipAddress.length() > 0 && ipAddress.endsWith("\n")) {
            ipAddress = ipAddress.substring(0, ipAddress.length() - 1);
        }
        LibvirtHost.logger.debug("Mac-to-IP " + macAddress + " -> " + ipAddress);
        return ipAddress;
    }

    @Override
    public ObjectName getObjectName() {
        return this.objectName;
    }

    private HashMap<String, String> hypervisorInfo;

    public Map<String, String> getHypervisorInfo() {
        if (this.hypervisorInfo == null) {
            this.hypervisorInfo = new HashMap<String, String>();
            try {
                this.hypervisorInfo.put("name", this.connection.getType());
                long v = this.connection.getVersion();
                long major = v / 1000000;
                long minor = (v - major * 1000000) / 1000;
                this.hypervisorInfo.put("version", "" + major + "." + minor);
            } catch (LibvirtException ex) {
                LibvirtHost.logger.error("Failed to get hypervisor info", ex);
            }
        }
        return this.hypervisorInfo;
    }

    private HashMap<String, String> cpuInfo;

    public Map<String, String> getCPUInfo() {
        if (this.cpuInfo == null) {
            this.cpuInfo = new HashMap<String, String>();
            NodeInfo nodeInfo;
            try {
                nodeInfo = this.connection.nodeInfo();
            } catch (LibvirtException e) {
                e.printStackTrace();
                return this.cpuInfo;
            }
            this.cpuInfo.put("model", new String(nodeInfo.model));
            this.cpuInfo.put("speedMHz", Integer.toString(nodeInfo.mhz));
            this.numCPU = nodeInfo.cpus;
        }
        return this.cpuInfo;
    }

    public int getNumCPU() {
        if (this.numCPU == -1) {
            this.getCPUInfo();
        }
        return this.numCPU;
    }

    public String getHostName() {
        return this.hostName;
    }

    public VirtualMachineImageStoreMXBean getVMImageStore() {
        return this.serverPool.getVMImageStore();
    }

    public long getFreeMemoryMB() {
        try {
            if (this.serverPool.getHypervisor().equals("xen")) {
                return this.getLibvirtFreeMemoryMB();
            } else if (this.serverPool.getHypervisor().equals("kvm")) {
                return this.getLinuxHostFreeMemoryMB();
            }
        } catch (Exception ex) {
            LibvirtHost.logger.error("getFreeMemoryMB", ex);
        }
        return -1;
    }

    private long getLibvirtFreeMemoryMB() throws Exception {
        long freeMemKB = -1;
        freeMemKB = this.connection.getFreeMemory();
        return freeMemKB / (1024 * 1024);
    }

    private long getLinuxHostFreeMemoryMB() throws Exception {
        RemoteExec.Result result;
        String command = "free -m | sed '3!d' | awk '{print $4}'";
        result = RemoteExec.commandAsRoot(this.hostName, this.authInfo, command);
        if (result.output.endsWith("\n")) {
            result.output = result.output.substring(0, result.output.length() - 1);
        }
        long osFreeMemMB = Long.parseLong(result.output);
        long vmTotalMemoryMB = 0;
        for (VirtualMachineMXBean vm : this.vmList) {
            vmTotalMemoryMB += vm.getMemorySizeMB();
        }
        long vmFreeMemMB = this.getTotalMemoryMB() - vmTotalMemoryMB;
        return Math.min(osFreeMemMB, vmFreeMemMB);
    }

    public void sync() {
        this.hostSynchronizer.sync();
    }

    private ResourcePartitionMXBean findVMResourcePartition(final Domain domain) throws Exception {
        String description = LibvirtConfigFileUtils.getDescription(domain.getXMLDesc(0));
        if (description == null || !description.startsWith("resourcePartition:")) {
            return this.serverPool;
        }
        int i = description.indexOf(" ");
        if (i != -1 && description.length() > i + 1) {
            String rpPath = description.substring(i + 1);
            if (this.serverPool.getPath().equals(rpPath)) {
                return this.serverPool;
            }
            ResourcePartitionMXBean rp = this.serverPool.getSubPartitionByPath(rpPath);
            if (rp == null) {
                LibvirtHost.logger.error("Cannot find resourcePartition with path=" + rpPath);
                return this.serverPool;
            }
            return rp;
        }
        return this.serverPool;
    }

    private synchronized boolean doSync(final Connect connection) {
        try {
            String vmNames[] = connection.listDefinedDomains();
            for (String vmName : vmNames) {
                if (this.vmCreationInProgress(vmName)) {
                    continue;
                }
                Domain domain;
                try {
                    domain = connection.domainLookupByName(vmName);
                } catch (LibvirtException ex) {
                    continue;
                }
                String uuid = domain.getUUIDString();

                ObjectName name = MBeanObjectNamer.makeVirtualMachineName(this.serverPool.getPath() + "/" + vmName, uuid);
                LibvirtVirtualMachine vm = this.lookUpVMByName(vmName);
                if (AgentCommon.getMBeanServer().isRegistered(name)) {
                    if (vm != null) {
                        vm.setSynchronized(true);
                    }
                    continue;
                }
                if (vm == null) {
                    vm = new LibvirtVirtualMachine(name, this, domain, null);
                    vm.setResourcePartition(this.findVMResourcePartition(domain));
                    this.addVM(vm);
                    AgentCommon.getMBeanServer().registerMBean(vm, name);
                    this.emitNotification(NotificationTypes.VM_ADD, "Created", name);
                }
            }

            int domIDs[] = connection.listDomains();
            for (int domID : domIDs) {
                Domain domain;
                try {
                    domain = connection.domainLookupByID(domID);
                } catch (LibvirtException ex) {
                    continue;
                }
                String uuid = domain.getUUIDString();
                String vmName = domain.getName();
                if (vmName.equals("Domain-0")) {
                    // uuid = this.hostName; // XXX
                    continue; // Exclude Domain-0 from the vm list
                }
                ObjectName name = MBeanObjectNamer.makeVirtualMachineName(this.serverPool.getPath() + "/" + vmName, uuid);
                LibvirtVirtualMachine vm = this.lookUpVMByName(vmName);
                if (AgentCommon.getMBeanServer().isRegistered(name)) {
                    if (vm != null) {
                        vm.setSynchronized(true);
                    }
                    continue;
                }
                if (vm == null) {
                    vm = new LibvirtVirtualMachine(name, this, domain, null);
                    vm.setResourcePartition(this.findVMResourcePartition(domain));
                    this.addVM(vm);
                    AgentCommon.getMBeanServer().registerMBean(vm, name);
                    this.emitNotification(NotificationTypes.VM_ADD, "Created", name);
                }
            }

            for (LibvirtVirtualMachine vm : this.vmList) {
                try {
                    if (vm.isMigrating()) {
                        continue;
                    }
                    if (!vm.isSynchronized()) {
                        vm.removeFromResourcePartition();
                        if (this.removeVM(vm)) {
                            this.emitNotification(NotificationTypes.VM_DEL, "Destroyed", vm.getObjectName());
                            AgentCommon.getMBeanServer().unregisterMBean(vm.getObjectName());
                        }
                    } else {
                        try {
                            vm.getState();
                        } catch (VMMException ignore) {
                        }
                    }
                } finally {
                    vm.setSynchronized(false);
                }
            }

        } catch (LibvirtException ex) {
            return false;
        } catch (Exception ex) {
            LibvirtHost.logger.error("error while syncing", ex);
        }

        try {
            System.gc();
            Thread.sleep(200);
            System.runFinalization();
            System.gc();
            Thread.sleep(200);
            System.runFinalization();
            System.gc();
            Thread.sleep(200);
            System.runFinalization();
        } catch (InterruptedException ex) {
        }

        return true;
    }

    public List<VirtualMachineMXBean> getResidentVMs() {
        List<VirtualMachineMXBean> result = new ArrayList<VirtualMachineMXBean>(this.vmList);
        return result;
    }

    @Override
    public VirtualMachineMXBean createVM(final VMConfigSpec vmSpec, final boolean sync) throws InsufficientResourcesException,
        InvalidVMConfigException, VMMException {
        throw new UnsupportedOperationException();
    }

    public VirtualMachineMXBean createVM(final VMConfigSpec vmSpecIn, final ResourcePartitionMXBean partition,
        final boolean sync) throws InvalidVMConfigException, VMMException {
        throw new UnsupportedOperationException();
    }

    @Override
    public VirtualMachineMXBean createVirtualMachine(final VirtualMachineConfigSpec vmConfigSpec,
        final CustomizationSpec customizationSpec, final boolean startVM, final boolean sync)
        throws InsufficientResourcesException, InvalidVMConfigException, VMMException {
        LibvirtVirtualMachine vm;
        ObjectName name;
        Domain domain;
        LibvirtDomainConfig domainConfig;

        this.checkVMConfigSpec(vmConfigSpec);
        try {
            domainConfig = LibvirtDomainConfig.newDomainConfig(this.serverPool.getHypervisor() + ".tmpl");
        } catch (Exception ex) {
            LibvirtHost.logger.error("Failed to load libvirt domain config template", ex);
            throw new VMMException(ex.getMessage());
        }
        domainConfig.setName(vmConfigSpec.getName());
        domainConfig.setMemoryMB((int) vmConfigSpec.getMemoryMB());
        domainConfig.setNumVCPUs(vmConfigSpec.getNumVCPUs());

        ResourcePartitionMXBean resourcePartition = vmConfigSpec.getPartition();
        if (resourcePartition == null) {
            resourcePartition = this.serverPool;
        }
        domainConfig.setDescription("resourcePartition: " + resourcePartition.getPath());

        String devicePrefix = domainConfig.getDiskDevicePrefix();
        char deviceSuffix = 'a';
        int deviceIndex = 0;

        for (VirtualDiskSpec diskSpec : vmConfigSpec.getDiskSpecs()) {
            Volume vol;
            LibvirtStoragePool storagePool = (LibvirtStoragePool) this.serverPool.getStoragePoolByName(diskSpec
                .getStoragePool().getName());
            if (storagePool == null) {
                LibvirtHost.logger.error("Cannot find libvirt storage pool " + diskSpec.getStoragePool().getName());
                continue;
            }
            String volName = "vol-" + vmConfigSpec.getName() + "-" + System.nanoTime();
            if (diskSpec.getDiskOp() == DiskOperation.NEW) {
                vol = storagePool.createVolume(volName, diskSpec.getCapacityMB(), this.diskImageFormat);
            } else if (diskSpec.getDiskOp() == DiskOperation.CREATE_FROM) {
                vol = storagePool.createVolumeFrom(volName, diskSpec.getCapacityMB(), this.diskImageFormat,
                    diskSpec.getVolume(), true);
                if (customizationSpec != null) {
                    try {
                        storagePool.customizeVolume(vol, customizationSpec);
                    } catch (Exception e) {
                        LibvirtHost.logger.error("Failed to customize disk " + vol.getPath(), e);
                    }
                }
            } else {
                vol = diskSpec.getVolume();
            }
            domainConfig.addDisk(vol.getPath(), devicePrefix + new String(new char[] {(char) (deviceSuffix + deviceIndex++)}));
        }
        if (vmConfigSpec.getCdromSpecs() != null) {
            if (!domainConfig.getDiskDevicePrefix().equals(domainConfig.getCdromDevicePrefix())) {
                devicePrefix = domainConfig.getCdromDevicePrefix();
                deviceSuffix = 'c';
                deviceIndex = 0;
            }

            for (VirtualCdromSpec cdromSpec : vmConfigSpec.getCdromSpecs()) {
                String path;
                if (cdromSpec.getIsoVolume() != null) {
                    path = cdromSpec.getIsoVolume().getPath();
                } else {
                    path = "";
                }
                domainConfig.addCdrom(path, devicePrefix + new String(new char[] {(char) (deviceSuffix + deviceIndex++)}));
            }
        }
        if (vmConfigSpec.getVnicSpecs() != null) {
            for (VNICSpec vnicSpec : vmConfigSpec.getVnicSpecs()) {
                if (vnicSpec.getAddressType().equals("manual")) {
                    domainConfig.addVNIC(vnicSpec.getMacAddress());
                } else {
                    domainConfig.addVNIC(null);
                }
            }
        }
        // if (vmConfigSpec.getProperties() != null) {
        // for (String prop : vmConfigSpec.getProperties().keySet()) {
        // if (prop.equals("bootDevice")) {
        // domainConfig.setBootDevice(vmConfigSpec.getProperties().get("bootDevice"));
        // }
        // }
        // }

        String domainXml = domainConfig.toString();
        LibvirtHost.logger.info("Defining domain:\n" + domainXml);

        try {
            domain = this.connection.domainDefineXML(domainXml);
        } catch (LibvirtException ex) {
            LibvirtHost.logger.error("Failed to define domain: ", ex);
            throw new VMMException(ex.getMessage());
        }
        synchronized (this) {
            try {
                domain.setAutostart(true);
                String vmUuid = domain.getUUIDString();
                LibvirtHost.logger.debug("Created VM uuid=" + vmUuid);
                // LibvirtHost.logger.debug("Create VM xml:" +
                // domain.getXMLDesc(0));
                vm = this.lookUpVMByUUID(vmUuid);
                if (vm == null) {
                    name = MBeanObjectNamer.makeVirtualMachineName(this.serverPool.getPath() + "/" + vmConfigSpec.getName(),
                        vmUuid);
                    vm = new LibvirtVirtualMachine(name, this, domain, null);
                    vm.setResourcePartition(resourcePartition);
                    this.addVM(vm);
                    AgentCommon.getMBeanServer().registerMBean(vm, name);
                    this.emitNotification(NotificationTypes.VM_ADD, "Created", name);
                }
            } catch (Exception ex) {
                LibvirtHost.logger.debug("Host " + this.hostName, ex);
                throw new VMMException(ex.getMessage());
            }
        }

        if (startVM) {
            vm.start();
        }

        return vm;
    }

    public VirtualMachineMXBean cloneVM(final String targetVmName, final String clonedVmName,
        final VMCustomizationSpec custSpec, final boolean sync) throws InsufficientResourcesException, VMMException {
        String command = "cloneVM --src " + clonedVmName + " --name " + targetVmName + " --force ";
        if (custSpec != null) {
            if (custSpec.getGuestOsHostName() != null && !custSpec.getGuestOsHostName().equals("")) {
                command += " --hostname " + custSpec.getGuestOsHostName();
            }
            // XXX configure only first NIC
            if (custSpec.getVnicIpSettingsList() != null && custSpec.getVnicIpSettingsList().size() > 0) {
                VnicIPSettings nic0Settings = custSpec.getVnicIpSettingsList().get(0);
                if (nic0Settings.getIpAssignmentMode() == IpAssignmentMode.DHCP) {
                    command += " --net eth0/dhcp";
                } else if (nic0Settings.getIpAssignmentMode() == IpAssignmentMode.FIXED) {
                    command += " --net eth0/static/" + nic0Settings.getIpAddress() + "/" + nic0Settings.getSubnetMask() + "/"
                        + nic0Settings.getGateway();
                }
            }
        }
        RemoteExec.Result result;
        try {
            result = RemoteExec.commandAsRoot(this.getHostName(), this.getSshAuthInfo(), command);
        } catch (RemoteExec.SshException ex) {
            throw new VMMException("VM cloning failed: SSH connection failure", ex);
        }
        if (result.exitCode != 0) {
            throw new VMMException("VM cloning failed: " + result.error);
        }

        Domain newVM = null;
        String vmUUID = "";

        try {
            newVM = this.connection.domainLookupByName(clonedVmName);
            vmUUID = newVM.getUUIDString();
        } catch (LibvirtException ex) {
            LibvirtHost.logger.error("Failed to look up VM " + clonedVmName);
            throw new VMMException("VM cloning failure");
        }

        synchronized (this) {
            ObjectName name = null;
            LibvirtVirtualMachine vm = null;
            vm = this.lookUpVMByUUID(vmUUID);
            if (vm != null) {
                return vm;
            }
            try {
                name = MBeanObjectNamer.makeVirtualMachineName(this.serverPool.getPath() + "/" + clonedVmName, vmUUID);
                HashMap<String, String> map = new HashMap<String, String>();
                vm = new LibvirtVirtualMachine(name, this, newVM, map);
                this.addVM(vm);

                AgentCommon.getMBeanServer().registerMBean(vm, name);
            } catch (Exception ex) {
                LibvirtHost.logger.debug("Host " + this.hostName, ex);
                throw new VMMException(ex);
            }

            this.emitNotification(NotificationTypes.VM_ADD, "Created", name);

            return vm;
        }
    }

    public long getTotalMemoryMB() {
        NodeInfo nodeInfo;
        try {
            nodeInfo = this.connection.nodeInfo();
        } catch (LibvirtException e) {
            e.printStackTrace();
            return -1;
        }
        return nodeInfo.memory / 1024;
    }

    // XX unsupported
    public float[] getLoadPerCPU() {
        float[] result = new float[0];
        return result;
    }

    public float getCPULoad() {
        float aggregatedCPULoad = 0;
        for (LibvirtVirtualMachine vm : this.vmList) {
            aggregatedCPULoad += vm.getResourceUsage().getCpuLoad();
        }
        return aggregatedCPULoad;
    }

    @Override
    public MBeanNotificationInfo[] getNotificationInfo() {
        return new MBeanNotificationInfo[] {new MBeanNotificationInfo(new String[] {NotificationTypes.VM_STATE_CHANGE,
            NotificationTypes.VM_ADD, NotificationTypes.VM_INVENTORY_DEL, NotificationTypes.VM_DEL, NotificationTypes.ERROR,
            NotificationTypes.LOG}, Notification.class.getName(), "Host event")};
    }

    public Map<String, ResourceUsage> getVMResourceUsage() {
        HashMap<String, ResourceUsage> result = new HashMap<String, ResourceUsage>();
        for (LibvirtVirtualMachine vm : this.vmList) {
            result.put(vm.getNameLabel(), vm.getResourceUsage());
        }
        return result;
    }

    public Map<String, Float> getVMCPULoads() {
        HashMap<String, Float> cpuLoads = new HashMap<String, Float>();
        for (LibvirtVirtualMachine vm : this.vmList) {
            cpuLoads.put(vm.getNameLabel(), vm.getResourceUsage().getCpuLoad());
        }
        return cpuLoads;
    }

    private class HostSynchronizer implements Runnable {
        private long periodMillis;

        private volatile boolean stopRequested;

        private Thread thread;

        private Connect connection;

        HostSynchronizer(final long periodMillis) {
            this.periodMillis = periodMillis;
        }

        public void requestStop() {
            this.stopRequested = true;
            this.sync();
        }

        void start() {
            this.thread = new Thread(this);
            this.stopRequested = false;
            this.thread.start();
        }

        synchronized void sync() {
            this.notify();
        }

        @Override
        public void run() {
            while (!this.stopRequested) {
                try {
                    this.connection = LibvirtHost.this.getLibvirtAPIConnection();
                    if (LibvirtHost.this.hostPowerState == HostPowerState.NOT_RESPONDING) {
                        LibvirtHost.this.setHostStatusFromNonRespondingToRunning();
                    } else {
                        if (LibvirtHost.this.doSync(this.connection) == false) {
                            LibvirtHost.this.setHostStatusFromRunningToNonResponding();
                            this.connection = null;
                        }
                    }
                } catch (LibvirtException ex) {
                    // LibvirtHost.logger.error("Failed to create Libvirt connection with "
                    // + LibvirtHost.this.uri, ex);
                    if (LibvirtHost.this.hostPowerState == HostPowerState.RUNNING) {
                        LibvirtHost.this.setHostStatusFromRunningToNonResponding();
                        this.connection = null;
                    }
                }

                synchronized (this) {
                    try {
                        this.wait(this.periodMillis);
                    } catch (InterruptedException ex) {
                    }
                }
            }
        }
    }

    @Override
    public Map<String, String> getAttributes() {
        return this.hostAttributes;
    }

    @Override
    public void release() {
        this.hostSynchronizer.requestStop();
        for (LibvirtVirtualMachine vm : this.vmList) {
            this.removeVMFromInventory(vm);
        }
        this.vmList.clear();
        try {
            if (this.connection != null) {
                this.connection.close();
            }
        } catch (LibvirtException ex) {
        }
        this.connection = null;
    }

    private void setHostStatusFromRunningToNonResponding() {
        while (this.getResidentVMs().size() > 0) {
            this.removeVMFromInventory((LibvirtVirtualMachine) this.getResidentVMs().get(0));
        }
        this.hostPowerState = HostPowerState.NOT_RESPONDING;
        this.connection = null;
        LibvirtHost.logger.warn("Host " + LibvirtHost.this.hostName + " not responding");
        this.emitNotification(NotificationTypes.HOST_STATE_CHANGE, this.hostPowerState.toString(), null);
    }

    private void setHostStatusFromNonRespondingToRunning() {
        this.hostPowerState = HostPowerState.RUNNING;
        this.emitNotification(NotificationTypes.HOST_STATE_CHANGE, this.hostPowerState.toString(), null);
    }

    @Override
    public void start() {
        try {
            this.connection = this.getLibvirtAPIConnection();
            this.hostSynchronizer.start();
        } catch (LibvirtException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        this.hostPowerState = HostPowerState.RUNNING;
        this.emitNotification(NotificationTypes.HOST_STATE_CHANGE, this.hostPowerState.toString(), null);
    }

    @Override
    public void stop() {
        this.hostSynchronizer.requestStop();

        while (this.getResidentVMs().size() > 0) {
            try {
                while (this.getResidentVMs().get(0).getState() == PowerState.RUNNING) {
                    this.getResidentVMs().get(0).shutdown();
                }
                this.removeVMFromInventory((LibvirtVirtualMachine) this.getResidentVMs().get(0));
            } catch (BadVMPowerStateException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            } catch (VMMException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();

            }
        }

        this.hostPowerState = HostPowerState.HALTED;
        this.emitNotification(NotificationTypes.HOST_STATE_CHANGE, this.hostPowerState.toString(), null);
    }

    /*
     * (non-Javadoc)
     * @see org.ow2.sirocco.vmm.api.HostMXBean#getHostState()
     */
    public HostPowerState getHostState() throws VMMException {
        return this.hostPowerState;
    }

}
