/**
 *
 * SIROCCO
 * Copyright (C) 2011 France Telecom
 * Contact: sirocco@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  $Id: LibvirtStoragePool.java 204 2011-07-27 10:08:27Z dangtran $
 *
 */

package org.ow2.sirocco.vmm.agent.driver.libvirt;

import java.io.ByteArrayInputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import javax.management.ObjectName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.apache.log4j.Logger;
import org.libvirt.Connect;
import org.libvirt.LibvirtException;
import org.libvirt.StoragePool;
import org.libvirt.StoragePoolInfo;
import org.libvirt.StorageVol;
import org.libvirt.StorageVolInfo;
import org.ow2.sirocco.vmm.agent.domain.AbstractStoragePool;
import org.ow2.sirocco.vmm.agent.driver.util.RemoteExec;
import org.ow2.sirocco.vmm.agent.driver.util.RemoteExec.SshException;
import org.ow2.sirocco.vmm.agent.main.AgentCommon;
import org.ow2.sirocco.vmm.api.CustomizationSpec;
import org.ow2.sirocco.vmm.api.InsufficientResourcesException;
import org.ow2.sirocco.vmm.api.StoragePoolMXBean;
import org.ow2.sirocco.vmm.api.VMMException;
import org.ow2.sirocco.vmm.api.VnicIPSettings;
import org.ow2.sirocco.vmm.api.VnicIPSettings.IpAssignmentMode;
import org.ow2.sirocco.vmm.api.Volume;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

public class LibvirtStoragePool extends AbstractStoragePool implements StoragePoolMXBean {
    static Logger logger = Logger.getLogger(LibvirtStoragePool.class);

    static final String RAW_FORMAT = "raw";

    static final String QCOW2_FORMAT = "qcow2";

    private String poolName;

    private Map<String, String> properties;

    private StoragePool storagePool;

    private String storagePoolPath;

    private LibvirtHost host;

    private Map<String, Volume> volumeMap = new HashMap<String, Volume>();

    LibvirtStoragePool(final ObjectName objectName, final String poolName, final LibvirtHost host,
        final Map<String, String> properties) throws Exception {
        super(objectName);
        this.poolName = poolName;
        this.host = host;
        Connect connection = host.getLibvirtAPIConnection();
        this.storagePool = connection.storagePoolLookupByName(poolName);
        Document doc = this.parseXMLDesc(this.storagePool.getXMLDesc(0));
        Node node = XMLUtils.getNodeByPath(doc, "/pool/target/path").item(0);
        this.storagePoolPath = node.getTextContent();
        LibvirtStoragePool.logger.debug("StoragePool " + poolName + " dir=" + this.storagePoolPath);
        if (this.storagePool.isActive() == 0) {
            this.storagePool.create(0);
        }
        this.populate();
        this.properties = new HashMap<String, String>(properties);
        LibvirtStoragePool.logger.debug("capacity=" + this.getCapacityMB() + "MB freeSpace=" + this.getFreeSpaceMB() + "MB");
    }

    @Override
    public void release() {
        try {
            if (this.storagePool != null) {
                this.storagePool.free();
            }
        } catch (LibvirtException e) {
        }
        this.storagePool = null;
    }

    @Override
    public String getName() {
        return this.poolName;
    }

    @Override
    public String[] getSupportedDiskImageFormats() {
        return new String[] {LibvirtStoragePool.QCOW2_FORMAT, LibvirtStoragePool.RAW_FORMAT};
    }

    @Override
    public Map<String, String> getProperties() {
        return this.properties;
    }

    String getStoragePoolPath() {
        return this.storagePoolPath;
    }

    @Override
    public void refresh() {
        try {
            this.storagePool.refresh(0);
            this.populate();
        } catch (Exception ex) {
            LibvirtStoragePool.logger.error("Failed to refresh storagePool " + this.poolName, ex);
        }
    }

    private synchronized void populate() throws Exception {
        this.volumeMap.clear();
        String[] volNames = this.storagePool.listVolumes();
        for (String volName : volNames) {
            StorageVol storageVol = this.storagePool.storageVolLookupByName(volName);
            Volume volume = this.newVolume(storageVol);

            LibvirtStoragePool.logger.debug("VOL=" + volume);

            this.volumeMap.put(volume.getKey(), volume);
        }
    }

    private Volume newVolume(final StorageVol storageVol) throws Exception {
        Volume volume = new Volume();
        volume.setStoragePool(this);
        volume.setKey(storageVol.getKey());
        volume.setName(storageVol.getName());
        volume.setPath(storageVol.getPath());
        StorageVolInfo volInfo = storageVol.getInfo();
        volume.setCapacityMB(volInfo.capacity / (1024 * 1024));
        volume.setAllocationMB(volInfo.allocation / (1024 * 1024));
        Document doc = this.parseXMLDesc(storageVol.getXMLDesc(0));
        Node node = XMLUtils.getNodeByPath(doc, "/volume/target").item(0);
        String format = XMLUtils.getAttributeValue(node, "format", "type");
        volume.setFormat(format);
        return volume;
    }

    private Document parseXMLDesc(final String xmlDesc) throws Exception {
        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        DocumentBuilder db = dbf.newDocumentBuilder();
        return db.parse(new ByteArrayInputStream(xmlDesc.getBytes("UTF-8")));
    }

    @Override
    public long getCapacityMB() throws VMMException {
        try {
            return this.storagePool.getInfo().capacity / (1024 * 1024);
        } catch (LibvirtException ex) {
            throw new VMMException(ex.toString());
        }
    }

    @Override
    public long getFreeSpaceMB() throws VMMException {
        try {
            StoragePoolInfo info = this.storagePool.getInfo();
            return (info.capacity - info.allocation) / (1024 * 1024);
        } catch (LibvirtException ex) {
            throw new VMMException(ex.toString());
        }
    }

    @Override
    public String getType() {
        return "file-backed"; // XXX
    }

    @Override
    public synchronized List<Volume> getVolumes() throws VMMException {
        return new ArrayList<Volume>(this.volumeMap.values());
    }

    @Override
    public Volume getVolumeByKey(final String key) {
        Volume vol = this.volumeMap.get(key);
        return vol;
    }

    @Override
    public void destroyVolume(final String key) throws VMMException {
        Volume vol = this.volumeMap.remove(key);
        if (vol == null) {
            throw new VMMException("Volume with key=" + key + " not found");
        }
        try {
            StorageVol storageVol = this.storagePool.storageVolLookupByName(vol.getName());
            if (storageVol != null) {
                storageVol.delete(0);
                LibvirtStoragePool.logger.info("Destroyed volume " + vol.getKey());
            } else {
                LibvirtStoragePool.logger.error("Failed to destroy volume " + vol.getName() + ": not found");
            }
        } catch (LibvirtException e) {
            LibvirtStoragePool.logger.error("While destroying volume " + vol.getPath(), e);
            throw new VMMException("Failed to destroy volume " + e.getMessage());
        }
    }

    private Volume createVolumeFromXMLDesc(final String volXmlDesc, final Volume parent) throws Exception {
        StorageVol storageVol;
        if (parent == null) {
            storageVol = this.storagePool.storageVolCreateXML(volXmlDesc, 0);
        } else {
            LibvirtStoragePool parentPool = (LibvirtStoragePool) this.host.serverPool.getStoragePoolByName(parent
                .getStoragePool().getName());
            StorageVol parentStorageVol = parentPool.storagePool.storageVolLookupByName(parent.getName());
            storageVol = this.storagePool.storageVolCreateXMLFrom(volXmlDesc, parentStorageVol, 0);
        }
        Volume volume = this.newVolume(storageVol);
        synchronized (this) {
            this.volumeMap.put(volume.getKey(), volume);
        }
        return volume;
    }

    @Override
    public Volume createVolume(final String name, final long capacityMB, final String format) throws VMMException {
        if (!format.equals(LibvirtStoragePool.RAW_FORMAT) && !format.equals(LibvirtStoragePool.QCOW2_FORMAT)) {
            throw new VMMException("Unsupported image format: " + format);
        }
        LibvirtStoragePool.logger.debug("Creating volume name=" + name + " capacityMB=" + capacityMB + " format=" + format);

        String volFileName = name + "-" + System.nanoTime() + "." + format;
        String volXmlDesc = "<volume> <name>" + volFileName + "</name> <allocation>0</allocation><capacity unit='M'>"
            + capacityMB + "</capacity><target> <format type='" + format + "'/></target></volume>";
        try {
            return this.createVolumeFromXMLDesc(volXmlDesc, null);
        } catch (Exception e) {
            throw new VMMException("Failed to create volume " + e.getMessage());
        }
    }

    @Override
    public Volume createVolumeFrom(final String name, final long capacityMB, final String format, final Volume parentVolume,
        final boolean copyOnWrite) throws VMMException {
        if (!format.equals(LibvirtStoragePool.RAW_FORMAT) && !format.equals(LibvirtStoragePool.QCOW2_FORMAT)) {
            throw new VMMException("Unsupported image format: " + format);
        }
        LibvirtStoragePool.logger.debug("Creating volume name=" + name + " capacityMB=" + capacityMB + " format=" + format
            + " parent=" + parentVolume.getKey());

        String volFileName = name + "-" + System.nanoTime() + "." + format;

        if (format.equals(LibvirtStoragePool.QCOW2_FORMAT)) {
            String volXmlDesc;
            if (copyOnWrite) {
                volXmlDesc = "<volume> <name>" + volFileName + "</name> <capacity unit='M'>" + parentVolume.getCapacityMB()
                    + "</capacity><target> <format type='" + format + "'/></target> <backingStore><path>"
                    + parentVolume.getPath() + "</path> <format type='" + format + "'/></backingStore> </volume>";
            } else {
                volXmlDesc = "<volume> <name>" + volFileName + "</name> <capacity unit='M'>" + capacityMB
                    + "</capacity><target> <format type='" + format + "'/> </target></volume>";
            }
            try {
                return this.createVolumeFromXMLDesc(volXmlDesc, copyOnWrite ? null : parentVolume);
            } catch (Exception e) {
                throw new VMMException("Failed to create volume " + e.getMessage());
            }
        } else {
            // libvirt vol-create-from does not handle sparse files, use unix cp
            // instead
            RemoteExec.Result result;
            try {
                String command = "cp " + parentVolume.getPath() + " " + this.storagePoolPath + "/" + volFileName;
                LibvirtStoragePool.logger.debug("Creating raw image file from source: " + command);
                result = RemoteExec.commandAsRoot(this.host.getHostName(), this.host.getSshAuthInfo(), command);
            } catch (SshException ex) {
                throw new VMMException("Failed to create volume: " + ex.getMessage());
            }
            if (result.exitCode != 0) {
                throw new VMMException("Failed to create volume:" + result.error);
            }
            Volume volume = new Volume();
            volume.setStoragePool(this);
            volume.setKey(this.storagePoolPath + "/" + volFileName);
            volume.setName(volFileName);
            volume.setPath(volume.getKey());
            volume.setCapacityMB(parentVolume.getCapacityMB());
            volume.setAllocationMB(0);
            volume.setFormat(LibvirtStoragePool.RAW_FORMAT);
            this.volumeMap.put(volume.getKey(), volume);
            try {
                this.storagePool.refresh(0);
            } catch (Exception ex) {
                LibvirtStoragePool.logger.error("Failed to refresh storagePool " + this.poolName, ex);
            }
            return volume;
        }
    }

    @Override
    public Volume uploadVolume(final String url, final String format, final String name, final String description,
        final String checkSum) throws InsufficientResourcesException, VMMException {
        RemoteExec.Result result;
        // XXX
        String volName = name + UUID.randomUUID().toString() + "." + format;
        String command = "wget -O " + this.storagePoolPath + "/" + volName + " " + url;
        try {
            result = RemoteExec.commandAsRoot(this.host.getHostName(), this.host.getSshAuthInfo(), command);
        } catch (SshException ex) {
            throw new VMMException(ex.getMessage());
        }
        if (result.exitCode != 0) {
            throw new VMMException(result.error);
        }
        try {
            this.storagePool.refresh(0);
            StorageVol storageVol = this.storagePool.storageVolLookupByName(volName);
            Volume volume = this.newVolume(storageVol);
            synchronized (this) {
                this.volumeMap.put(volume.getKey(), volume);
            }
            return volume;
        } catch (Exception e) {
            throw new VMMException("Failed to upload volume " + e.getMessage());
        }
    }

    void customizeVolume(final Volume volume, final CustomizationSpec customizationSpec) throws Exception {
        RemoteExec.Result result;
        String script = AgentCommon.getScript(LibvirtDriver.NAME, LibvirtDriver.CUSTOMIZE_DISK_SCRIPT);
        StringBuilder command = new StringBuilder("bash -s ");
        command.append("disk " + volume.getPath() + " ");
        if (customizationSpec.getGuestOsHostName() != null && !customizationSpec.getGuestOsHostName().equals("")) {
            command.append(" hostname " + customizationSpec.getGuestOsHostName());
        }
        // XXX configure only first NIC
        if (customizationSpec.getVnicIpSettingsList() != null && customizationSpec.getVnicIpSettingsList().size() > 0) {
            VnicIPSettings nic0Settings = customizationSpec.getVnicIpSettingsList().get(0);
            if (nic0Settings.getIpAssignmentMode() == IpAssignmentMode.DHCP) {
                command.append("  net eth0/dhcp");
            } else if (nic0Settings.getIpAssignmentMode() == IpAssignmentMode.FIXED) {
                command.append("  net eth0/static/" + nic0Settings.getIpAddress() + "/" + nic0Settings.getSubnetMask() + "/"
                    + nic0Settings.getGateway());
            }
        }
        if (customizationSpec.getGuestOsCustomizationParams() != null) {
            for (String param : customizationSpec.getGuestOsCustomizationParams().keySet()) {
                if (param.equals("sshkey")) {
                    command.append(" sshkey \"" + customizationSpec.getGuestOsCustomizationParams().get(param) + "\"");
                } else {
                    command.append(" userdata \"" + param + "\" \""
                        + customizationSpec.getGuestOsCustomizationParams().get(param) + "\"");
                }
            }
        }
        command.append(" << 'END'\n");
        command.append(script);
        command.append("\n");
        command.append("END");
        LibvirtStoragePool.logger.debug("Customizing disk " + volume.getPath() + "\n" + command.toString());
        result = RemoteExec.commandAsRoot(this.host.getHostName(), this.host.getSshAuthInfo(), command.toString());
        if (result.exitCode != 0) {
            LibvirtStoragePool.logger.error("Failed to customize disk image: " + result.error);
            throw new Exception(result.error);
        }
    }
}
